@@ -11,7 +11,7 @@ import pandas as pd
 from Bio import SeqIO
 from Bio.Seq import Seq
 from Bio.SeqRecord import SeqRecord
-from pychangepoints.algo_changepoints import pelt
+from pychangepoints.algo_changepoints import pelt, binseg
 from skbio import read
 from skbio.alignment import TabularMSA
 from skbio.sequence import Protein
@@ -67,7 +67,7 @@ def pelt_ordered_indices(data, min_seg, pen=1e-3):
     min_seg : int
         Minimum segment length (number of observations between changes)
     pen : float
-        TODO
+        Penalty term applied to the PELT procedure
     """
     if pen < 1e-3:
         logger.warning(
@@ -77,6 +77,25 @@ def pelt_ordered_indices(data, min_seg, pen=1e-3):
     return np.sort(indices) - 1
 
 
+def binseg_ordered_indices(data, min_seg, nb_changepoints=10):
+    """Produce segmentation indices using BINSEG algorithm.
+
+    See more details on the related R documentation of this method:
+    https://www.rdocumentation.org/packages/changepoint/versions/2.2.2/topics/BINSEG.
+
+    Parameters
+    ----------
+    data : pd.DataFrame
+        Data on which BINSEG algorithm must be run to find changepoints
+    min_seg : int
+        Minimum segment length (number of observations between changes)
+    Q : int
+        Maximal number of changepoints found by the algorithm
+    """
+    indices = binseg(data, nb_changepoints, min_seg, "mll_mean")
+    return np.sort(indices) - 1
+
+
 def compute_mean_segmentation(vector, indices):
     """Aggregate the data contained in vector following the provided indices sequence. The resulting
     structure is a numpy array containing mean values of subsequence of the input vector.
@@ -141,7 +160,7 @@ class WellCorrelation:
         depth_column="MD",
         well_column="API",
         min_seg=1,
-        pelt_sup=0.2,
+        nb_markers=10,
         depth_min=10,
         depth_max=200,
         value_min=0,
@@ -181,8 +200,8 @@ class WellCorrelation:
             Positive integer giving the minimum segment length (number of observations between
             changes) ; used in PELT algorithm if ``segmentize_with_pelt`` is True, in a linear
             aggregation procedure otherwise
-        pelt_sup : float
-            Penalty term applied to the PELT procedure, when computing optimal consensus indices
+        nb_markers : int
+            Maximal number of markers requested in the consensus.
         quantile_list : tuple
             Global quantiles used in order to discriminate log dataseries. By default, one adds
             0.01 and 0.975 as extrem quantile values. This sequence must contain 22 values at most,
@@ -200,7 +219,7 @@ class WellCorrelation:
         self._match_column = match_column
         self._depth_column = depth_column
         self._min_seg = min_seg
-        self._pelt_sup = pelt_sup
+        self._nb_markers = nb_markers
         self._quantiles_list = quantile_list
         self._lr_normalize = lr_normalize
 
@@ -214,15 +233,11 @@ class WellCorrelation:
         # Well log multi-sequence alignment, as a skbio.TabularMSA
         self._multi_sequence_alignment = TabularMSA("")
         # Dictionnary matching the letter in the alignment with the depth for each well
-        self._depth_match_global = {}
+        self.depth_match_global = pd.DataFrame({"hole_id": [], "from": []})
         # Consensus
         self.consensus = pd.DataFrame(
             {self._match_column: [], self._depth_column: [], "freq": []}
         )
-        # Markers
-        self.markers = pd.DataFrame(
-            {self._depth_column: [], "marker_name": [], "hole_id": []}
-        )
 
         # For each well of the site of interest, compute normalized version of data series
         # The normalization is done by considering local mean values, depending on:
@@ -230,7 +245,7 @@ class WellCorrelation:
         # - or on a fixed step (simple version), see min_seg attribute
         self._dict_data = {}
         self._dict_data_original = {}
-        self._dict_segmentation = {}
+        self._segmentation_indices = {}
         for idx_well, well_data in data_log.groupby(well_column):
 
             well_name = str(well_data[well_column].iloc[0])
@@ -239,14 +254,19 @@ class WellCorrelation:
                 (depth_min < well_data.index)
                 & (well_data.index < depth_max)
                 & (value_min < well_data[self._match_column])
-                & (well_data[self._match_column] < value_max)
             ].copy()
             well_data.dropna(inplace=True)
+            if well_data.shape[0] <= 2:
+                logger.info("Not enough data for %s, skipping this site.", well_name)
+                continue
+            well_data.loc[
+                well_data[self._match_column] > value_max, self._match_column
+            ] = value_max
             # Depending on log_normalize parameter, consider logarithmic values
             if log_normalize:
                 well_data[self._match_column] = np.log(well_data[self._match_column])
             # Define the raw data series for this well
-            self._dict_data_original[well_name] = well_data
+            self._dict_data_original[well_name] = well_data.drop(well_column, axis=1)
 
             # If pelt is True, this algorithm is run to extract segmentation sequence
             if segmentize_with_pelt:
@@ -254,14 +274,14 @@ class WellCorrelation:
                     preprocessing.scale(well_data[[self._match_column]])
                 )
                 seg_match_small = pelt_ordered_indices(
-                    scaled_data, self.__PENALTY_PELT, self._min_seg
+                    scaled_data, self._min_seg, self.__PENALTY_PELT
                 )
             # Otherwise, one uses a regular sampling sequence
             else:
                 seg_match_small = np.arange(
-                    0, len(well_data[[self._match_column]]), self._min_seg
-                ).astype(int)
-            self._dict_segmentation[well_name] = seg_match_small
+                    0, well_data[self._match_column].shape[0], self._min_seg
+                )
+            self._segmentation_indices[well_name] = seg_match_small
 
             # Compute the normalized version of dataseries, knowing the segmentation sequence
             mean_query = compute_mean_segmentation(
@@ -293,7 +313,6 @@ class WellCorrelation:
         self.run_famsa()
         self.compute_depth_match()
         self.compute_consensus()
-        self.compute_markers(self._marker_index_global)
 
     def prepare_logs(self):
         """Prepare data logs: starting from segmented version of dataseries for each well, generate letter
@@ -405,7 +424,6 @@ class WellCorrelation:
 
         """
         # Set depth match global for each well
-        self._df_align = {}
         for key, data in self._dict_data_original.items():
             # Extract DNA sequences as strings for each well
             sequence = [
@@ -414,26 +432,31 @@ class WellCorrelation:
                 if protein.metadata["id"] == key
             ]
             assert len(sequence) == 1
-            global_align = np.array([item for item in sequence[0]])
-            # Produce a DataFrame with the sequence, and generate smart indices
-            df_align = pd.DataFrame({"letter": global_align})
+            # Produce a DataFrame with the sequence
+            df_align = pd.DataFrame({"letter": [item for item in sequence[0]]})
+            # Generate letter indices depending on the letters (as many indices as defined letters)
             mask = df_align["letter"] != "-"
             df_align.loc[mask, "indices"] = np.arange(df_align[mask].shape[0])
             df_align.loc[0, "indices"] = 0
             df_align["indices"] = (
                 df_align["indices"].fillna(method="ffill").astype("int")
             )
-            self._df_align[key] = df_align["indices"].values
-            # Here one gets non-empty letter indices,
-            # one builds depth_match_global item starting from raw data
-            seg_match = self._dict_segmentation[key][df_align["indices"].values]
-            self._depth_match_global[key] = np.array(data.index[seg_match])
+            # Here one gets non-empty letter indices...
+            # Recover segmentation indices that match aligned sequence indices
+            seg_match = self._segmentation_indices[key][df_align["indices"]]
+            # Build depth_match_global item starting from raw data
+            cur_df = pd.DataFrame(
+                {"hole_id": key, "from": data.iloc[seg_match].index.values}
+            )
+            self.depth_match_global = pd.concat([self.depth_match_global, cur_df])
 
     def compute_consensus(self):
         """Compute the consensus log, which is a "mean" log that aggregates the logs from every well. This
-        method also sets the value of marker_index_global, which represents a raw marker set
+        method also sets the value of consensus_depth, which corresponds to the depth of potential
+        markers in the consensus log generated by the instance of WellCorrelation.
 
         This method calls the PELT algorithm, developed in pychangepoints dependency.
+
         """
         # Prepare a DataFrame version of the alignment, with wells as indices
         df_msa = pd.DataFrame(self._multi_sequence_alignment.to_dict())
@@ -443,19 +466,23 @@ class WellCorrelation:
         ]
         # Compute letter frequency, for each position of the dataseries
         freq_df = df_msa.apply(pd.value_counts, axis=1)
-        letter_frequency = freq_df[self._vect_alphabet].sum(axis=1)
+        filtered_letters = np.intersect1d(self._vect_alphabet, freq_df.columns)
+        letter_frequency = freq_df[filtered_letters].sum(axis=1)
         # Building consensus series: an index belongs to the consensus
         # if there is data on a significative (see __CONSENSUS_per attribute) part of wells
         mask_consensus = letter_frequency > (1 - self.__CONSENSUS_per) * df_msa.shape[1]
+        filtered_alphabet = self._value_alphabet.loc[
+            self._value_alphabet.index.intersection(freq_df.columns.values), "quantile"
+        ]
         consensus = (
-            freq_df.drop(columns="-").fillna(0).dot(self._value_alphabet["quantile"])
+            freq_df.drop(columns="-").fillna(0).dot(filtered_alphabet)
             / letter_frequency
         )
         self.consensus = pd.DataFrame(
             {
                 self._match_column: np.repeat(consensus[mask_consensus], self._min_seg),
                 self._depth_column: np.linspace(
-                    0, mask_consensus.sum() - 1, mask_consensus.sum()
+                    0, mask_consensus.sum() - 1, mask_consensus.sum() * self._min_seg
                 )
                 / 10,
                 "freq": np.repeat(
@@ -463,64 +490,61 @@ class WellCorrelation:
                 ),
             }
         )
-        self.consensus.set_index(self._depth_column, inplace=True)
-        self._index_global_consensus = consensus[mask_consensus].index.values
+        self.consensus.index.rename("global_index", inplace=True)
+        self.consensus.reset_index(inplace=True)
         # Segmentize the consensus log to build the chronicle of aligned tops
         scaled_data = pd.DataFrame(
             preprocessing.scale(self.consensus[self._match_column])
         )
-        seg_signal = pelt_ordered_indices(scaled_data, 2, self._pelt_sup)
-        self.marker_index_in_consensus = seg_signal // self._min_seg
-        self._marker_index_global = np.unique(
-            self._index_global_consensus[self.marker_index_in_consensus]
-        ).astype(int)
+        seg_signal = binseg_ordered_indices(scaled_data, 2, self._nb_markers)
+        self.consensus_depth = self.consensus.iloc[seg_signal // self._min_seg][
+            self._depth_column
+        ]
 
-    def extract_marker_indices(self, tops):
-        """Given a list of depths, compute the corresponding indices in the consensus log. This function
-        aims at providing a valid input parameter for ``compute_markers`` method.
 
-        If the input list contains values that are not in consensus depth axis, these values are
-        not considered.
+def compute_markers(tops, df_aligned_depths, consensus, depth_column):
+    """Compute the markers associated to provided depths, for the well set defined by the given
+    consensus log and aligned depths dataseries.
 
-        Parameters
-        ----------
-        tops : list
-            Depths on the consensus log for which markers have to be expanded on the regular well
-            logs
+    Parameters
+    ----------
+    tops : list
+        Depths on the consensus log for which markers have to be expanded on the regular well
+        logs
+    df_aligned_depths : pd.DataFrame
+        Extended depth dataseries for all sites, aligned through FAMSA process or any alignment
+        algorithm
+    consensus : pd.DataFrame
+        Consensus log
+    depth_column : str
+        Name of the depth column in the consensus
 
-        """
-        mask = self.consensus.index.isin(tops)
-        idx_consensus = self.consensus.reset_index().loc[mask].index
-        return self._index_global_consensus[idx_consensus]
-
-    def compute_markers(self, aligned_positions):
-        """Compute the final state of markers associated to the well set
-
-        Parameters
-        ----------
-        aligned_positions : np.array
-            Index on the consensus log taken as references for marker generation
-        """
-        # Finding the consensus marker for each well
-        list_tops = []
-        for key in self._dict_data:
-            tops_current = self._depth_match_global[key][aligned_positions]
-            # Filter duplicated values at the beginning of the marker chronicle
-            m_tot = len(tops_current) - 1
-            i_j = 1
-            while (i_j < m_tot) & (tops_current[i_j] == tops_current[0]):
-                i_j += 1
-            # Filter duplicated values at the end of the marker chronicle
-            j_i = m_tot - 1
-            while (j_i > 1) & (tops_current[j_i] == tops_current[m_tot]):
-                j_i = j_i - 1
-            df_current = pd.DataFrame(
-                {
-                    "hole_id": key,
-                    "from": tops_current[(i_j - 1) : (j_i + 2)],
-                    "to": tops_current[(i_j - 1) : (j_i + 2)] + 0.1,
-                    "code": np.arange(i_j - 1, j_i + 2),
-                }
-            )
-            list_tops.append(df_current)
-        self.markers = pd.concat(list_tops, ignore_index=True)
+    Returns
+    -------
+    pd.DataFrame
+        Markers that correspond to the given tops, in the context of the provided consensus.
+    """
+    aligned_positions = consensus.loc[consensus[depth_column].isin(tops)][
+        "global_index"
+    ]
+    # Finding the consensus marker for each well
+    markers = pd.DataFrame({"hole_id": [], "from": [], "to": [], "code": []})
+    for key, group in df_aligned_depths.groupby("hole_id"):
+        tops_current = group.loc[aligned_positions].copy().reset_index(drop=True)
+        if tops_current.shape[0] == 0:
+            continue
+        tops_current.index.rename("code", inplace=True)
+        # Filter duplicated values at the beginning and at the end of the marker chronicle
+        begidx = tops_current[
+            tops_current["from"] == tops_current.iloc[0]["from"]
+        ].last_valid_index()
+        endidx = tops_current[
+            tops_current["from"] == tops_current.iloc[-1]["from"]
+        ].first_valid_index()
+        tops_current = tops_current.truncate(begidx, endidx)
+        # Fill in the resulting marker dataframe
+        tops_current["to"] = tops_current["from"] + 0.1
+        tops_current = tops_current.loc[tops_current.truncate(begidx, endidx).index]
+        markers = pd.concat([markers, tops_current.reset_index()])
+    markers["code"] = markers["code"].astype(int)
+    return markers
